package com.sun.jmx.remote.internal;

import java.rmi.UnmarshalException;
import java.io.NotSerializableException;
import javax.management.Notification;
import javax.management.remote.TargetedNotification;
import javax.management.MBeanServerNotification;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import javax.security.auth.Subject;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.management.ListenerNotFoundException;
import javax.management.InstanceNotFoundException;
import java.io.IOException;
import javax.management.remote.NotificationResult;
import java.security.AccessController;
import com.sun.jmx.remote.util.EnvHelp;
import java.util.HashMap;
import com.sun.jmx.remote.util.ClassLogger;
import java.util.Map;
import java.util.concurrent.Executor;
import java.security.AccessControlContext;

public abstract class ClientNotifForwarder
{
    private final AccessControlContext acc;
    private static int threadId;
    private final ClassLoader defaultClassLoader;
    private final Executor executor;
    private final Map<Integer, ClientListenerInfo> infoList;
    private long clientSequenceNumber;
    private final int maxNotifications;
    private final long timeout;
    private Integer mbeanRemovedNotifID;
    private Thread currentFetchThread;
    private static final int STARTING = 0;
    private static final int STARTED = 1;
    private static final int STOPPING = 2;
    private static final int STOPPED = 3;
    private static final int TERMINATED = 4;
    private int state;
    private boolean beingReconnected;
    private static final ClassLogger logger;
    
    public ClientNotifForwarder(Map map) {
        this(null, map);
    }
    
    public ClientNotifForwarder(ClassLoader defaultClassLoader, Map<String, ?> map) {
        this.infoList = new HashMap();
        this.clientSequenceNumber = -1L;
        this.mbeanRemovedNotifID = null;
        this.state = 3;
        this.beingReconnected = false;
        this.maxNotifications = EnvHelp.getMaxFetchNotifNumber(map);
        this.timeout = EnvHelp.getFetchTimeout(map);
        Executor executor = (Executor)map.get("jmx.remote.x.fetch.notifications.executor");
        if (executor == null) {
            executor = new ClientNotifForwarder.LinearExecutor();
        }
        else if (ClientNotifForwarder.logger.traceOn()) {
            ClientNotifForwarder.logger.trace("ClientNotifForwarder", "executor is " + executor);
        }
        this.defaultClassLoader = defaultClassLoader;
        this.executor = executor;
        this.acc = AccessController.getContext();
    }
    
    protected abstract NotificationResult fetchNotifs(long p0, int p1, long p2) throws IOException, ClassNotFoundException;
    
    protected abstract Integer addListenerForMBeanRemovedNotif() throws IOException, InstanceNotFoundException;
    
    protected abstract void removeListenerForMBeanRemovedNotif(Integer p0) throws IOException, InstanceNotFoundException, ListenerNotFoundException;
    
    protected abstract void lostNotifs(String p0, long p1);
    
    public synchronized void addNotificationListener(Integer n, ObjectName obj, NotificationListener obj2, NotificationFilter notificationFilter, Object o, Subject subject) throws IOException, InstanceNotFoundException {
        if (ClientNotifForwarder.logger.traceOn()) {
            ClientNotifForwarder.logger.trace("addNotificationListener", "Add the listener " + obj2 + " at " + obj);
        }
        this.infoList.put(n, new ClientListenerInfo(n, obj, obj2, notificationFilter, o, subject));
        this.init(false);
    }
    
    public synchronized Integer[] removeNotificationListener(ObjectName obj, NotificationListener obj2) throws ListenerNotFoundException, IOException {
        this.beforeRemove();
        if (ClientNotifForwarder.logger.traceOn()) {
            ClientNotifForwarder.logger.trace("removeNotificationListener", "Remove the listener " + obj2 + " from " + obj);
        }
        ArrayList<Integer> list = new ArrayList();
        ArrayList list2 = new ArrayList(this.infoList.values());
        for (int i = list2.size() - 1; i >= 0; --i) {
            ClientListenerInfo clientListenerInfo = (ClientListenerInfo)list2.get(i);
            if (clientListenerInfo.sameAs(obj, obj2)) {
                list.add(clientListenerInfo.getListenerID());
                this.infoList.remove(clientListenerInfo.getListenerID());
            }
        }
        if (list.isEmpty()) {
            throw new ListenerNotFoundException("Listener not found");
        }
        return (Integer[])list.toArray(new Integer[0]);
    }
    
    public synchronized Integer removeNotificationListener(ObjectName obj, NotificationListener obj2, NotificationFilter notificationFilter, Object o) throws ListenerNotFoundException, IOException {
        if (ClientNotifForwarder.logger.traceOn()) {
            ClientNotifForwarder.logger.trace("removeNotificationListener", "Remove the listener " + obj2 + " from " + obj);
        }
        this.beforeRemove();
        Integer listenerID = null;
        ArrayList list = new ArrayList(this.infoList.values());
        int i = list.size() - 1;
        while (i >= 0) {
            ClientListenerInfo clientListenerInfo = (ClientListenerInfo)list.get(i);
            if (clientListenerInfo.sameAs(obj, obj2, notificationFilter, o)) {
                listenerID = clientListenerInfo.getListenerID();
                this.infoList.remove(listenerID);
                break;
            }
            --i;
        }
        if (listenerID == null) {
            throw new ListenerNotFoundException("Listener not found");
        }
        return listenerID;
    }
    
    public synchronized Integer[] removeNotificationListener(ObjectName obj) {
        if (ClientNotifForwarder.logger.traceOn()) {
            ClientNotifForwarder.logger.trace("removeNotificationListener", "Remove all listeners registered at " + obj);
        }
        ArrayList<Integer> list = new ArrayList();
        ArrayList list2 = new ArrayList(this.infoList.values());
        for (int i = list2.size() - 1; i >= 0; --i) {
            ClientListenerInfo clientListenerInfo = (ClientListenerInfo)list2.get(i);
            if (clientListenerInfo.sameAs(obj)) {
                list.add(clientListenerInfo.getListenerID());
                this.infoList.remove(clientListenerInfo.getListenerID());
            }
        }
        return (Integer[])list.toArray(new Integer[0]);
    }
    
    public synchronized ClientListenerInfo[] preReconnection() throws IOException {
        if (this.state == 4 || this.beingReconnected) {
            throw new IOException("Illegal state.");
        }
        ClientListenerInfo[] array = (ClientListenerInfo[])this.infoList.values().toArray(new ClientListenerInfo[0]);
        this.beingReconnected = true;
        this.infoList.clear();
        return array;
    }
    
    public synchronized void postReconnection(ClientListenerInfo[] array) throws IOException {
        if (this.state == 4) {
            return;
        }
        while (this.state == 2) {
            try {
                this.wait();
            }
            catch (InterruptedException cause) {
                IOException throwable = new IOException(cause.toString());
                EnvHelp.initCause(throwable, cause);
                throw throwable;
            }
        }
        int traceOn = ClientNotifForwarder.logger.traceOn();
        for (int length = array.length, i = 0; i < length; ++i) {
            if (!(traceOn == 0)) {
                ClientNotifForwarder.logger.trace("addNotificationListeners", "Add a listener at " + array[i].getListenerID());
            }
            this.infoList.put(array[i].getListenerID(), array[i]);
        }
        this.beingReconnected = false;
        this.notifyAll();
        if (this.currentFetchThread != Thread.currentThread() && this.state != 0 && this.state != 1) {
            while (this.state == 2) {
                try {
                    this.wait();
                }
                catch (InterruptedException cause2) {
                    IOException throwable2 = new IOException(cause2.toString());
                    EnvHelp.initCause(throwable2, cause2);
                    throw throwable2;
                }
            }
            if (array.length > 0) {
                this.init(true);
                return;
            }
            if (this.infoList.size() > 0) {
                this.init(false);
                return;
            }
            return;
        }
        try {
            this.mbeanRemovedNotifID = this.addListenerForMBeanRemovedNotif();
        }
        catch (Exception t) {
            if (ClientNotifForwarder.logger.traceOn()) {
                ClientNotifForwarder.logger.trace("init", "Failed to register a listener to the mbean server: the client will not do clean when an MBean is unregistered", t);
            }
        }
        return;
    }
    
    public synchronized void terminate() {
        if (this.state == 4) {
            return;
        }
        if (ClientNotifForwarder.logger.traceOn()) {
            ClientNotifForwarder.logger.trace("terminate", "Terminating...");
        }
        if (this.state == 1) {
            this.infoList.clear();
        }
        this.setState(4);
        return;
    }
    
    private synchronized void setState(int state) {
        if (this.state == 4) {
            return;
        }
        this.state = state;
        this.notifyAll();
        return;
    }
    
    private synchronized void init(boolean b) throws IOException {
        switch (this.state) {
            case 1: {
                return;
            }
            case 0: {
                return;
            }
            case 4: {
                throw new IOException("The ClientNotifForwarder has been terminated.");
            }
            case 2: {
                if (this.beingReconnected) {
                    return;
                }
                while (this.state == 2) {
                    try {
                        this.wait();
                    }
                    catch (InterruptedException cause) {
                        IOException throwable = new IOException(cause.toString());
                        EnvHelp.initCause(throwable, cause);
                        throw throwable;
                    }
                }
                this.init(b);
                return;
            }
            case 3: {
                if (this.beingReconnected) {
                    return;
                }
                if (ClientNotifForwarder.logger.traceOn()) {
                    ClientNotifForwarder.logger.trace("init", "Initializing...");
                }
                if (!b) {
                    try {
                        NotificationResult fetchNotifs = this.fetchNotifs(-1L, 0, 0L);
                        if (this.state != 3) {
                            return;
                        }
                        this.clientSequenceNumber = fetchNotifs.getNextSequenceNumber();
                    }
                    catch (ClassNotFoundException ex) {
                        ClientNotifForwarder.logger.warning("init", "Impossible exception: " + ex);
                        ClientNotifForwarder.logger.debug("init", ex);
                    }
                }
                try {
                    this.mbeanRemovedNotifID = this.addListenerForMBeanRemovedNotif();
                }
                catch (Exception t) {
                    if (ClientNotifForwarder.logger.traceOn()) {
                        ClientNotifForwarder.logger.trace("init", "Failed to register a listener to the mbean server: the client will not do clean when an MBean is unregistered", t);
                    }
                }
                this.setState(0);
                this.executor.execute(new ClientNotifForwarder.NotifFetcher());
                return;
            }
            default: {
                throw new IOException("Unknown state.");
            }
        }
    }
    
    private synchronized void beforeRemove() throws IOException {
        while (this.beingReconnected) {
            if (this.state == 4) {
                throw new IOException("Terminated.");
            }
            try {
                this.wait();
            }
            catch (InterruptedException cause) {
                IOException throwable = new IOException(cause.toString());
                EnvHelp.initCause(throwable, cause);
                throw throwable;
            }
        }
        if (this.state == 4) {
            throw new IOException("Terminated.");
        }
        return;
    }
    
    static void access$1000(ClientNotifForwarder clientNotifForwarder, int state) {
        clientNotifForwarder.setState(state);
    }
    
    static {
        logger = new ClassLogger("javax.management.remote.misc", "ClientNotifForwarder");
    }
    
    private static class LinearExecutor implements Executor
    {
        private Runnable command;
        private Thread thread;
        
        @Override
        public synchronized void execute(Runnable command) {
            if (this.command != null) {
                throw new IllegalArgumentException("More than one command");
            }
            this.command = command;
            if (this.thread == null) {
                (this.thread = new Thread() {
                    @Override
                    public void run() {
                        while (true) {
                            Runnable access$000;
                            synchronized (ClientNotifForwarder.LinearExecutor.this) {
                                if (ClientNotifForwarder.LinearExecutor.this.command == null) {
                                    ClientNotifForwarder.LinearExecutor.this.thread = null;
                                    return;
                                }
                                access$000 = ClientNotifForwarder.LinearExecutor.this.command;
                                ClientNotifForwarder.LinearExecutor.this.command = null;
                            }
                            access$000.run();
                        }
                    }
                }).setDaemon(true);
                this.thread.setName("ClientNotifForwarder-" + ++ClientNotifForwarder.threadId);
                this.thread.start();
            }
            return;
        }
    }
    
    private class NotifFetcher implements Runnable
    {
        private volatile boolean alreadyLogged;
        
        private NotifFetcher() {
            this.alreadyLogged = false;
        }
        
        private void logOnce(String msg, SecurityException t) {
            if (this.alreadyLogged) {
                return;
            }
            ClientNotifForwarder.logger.config("setContextClassLoader", msg);
            if (t != null) {
                ClientNotifForwarder.logger.fine("setContextClassLoader", t);
            }
            this.alreadyLogged = true;
            return;
        }
        
        private final ClassLoader setContextClassLoader(ClassLoader classLoader) {
            AccessControlContext access$500 = ClientNotifForwarder.this.acc;
            if (access$500 == null) {
                this.logOnce("AccessControlContext must not be null.", null);
                throw new SecurityException("AccessControlContext must not be null");
            }
            return (ClassLoader)AccessController.doPrivileged(new PrivilegedAction<ClassLoader>() {
                @Override
                public ClassLoader run() {
                    try {
                        ClassLoader contextClassLoader = Thread.currentThread().getContextClassLoader();
                        if (classLoader == contextClassLoader) {
                            return contextClassLoader;
                        }
                        Thread.currentThread().setContextClassLoader(classLoader);
                        return contextClassLoader;
                    }
                    catch (SecurityException obj) {
                        ClientNotifForwarder.NotifFetcher.access$600(ClientNotifForwarder.NotifFetcher.this, "Permission to set ContextClassLoader missing. Notifications will not be dispatched. Please check your Java policy configuration: " + obj, obj);
                        throw obj;
                    }
                }
            }, access$500);
        }
        
        @Override
        public void run() {
            ClassLoader setContextClassLoader;
            if (ClientNotifForwarder.this.defaultClassLoader != null) {
                setContextClassLoader = this.setContextClassLoader(ClientNotifForwarder.this.defaultClassLoader);
            }
            else {
                setContextClassLoader = null;
            }
            try {
                this.doRun();
            }
            finally {
                if (ClientNotifForwarder.this.defaultClassLoader != null) {
                    this.setContextClassLoader(setContextClassLoader);
                }
            }
        }
        
        private void doRun() {
            synchronized (ClientNotifForwarder.this) {
                ClientNotifForwarder.this.currentFetchThread = Thread.currentThread();
                if (ClientNotifForwarder.this.state == 0) {
                    ClientNotifForwarder.access$1000(ClientNotifForwarder.this, 1);
                }
            }
            NotificationResult fetchNotifs = null;
            if (!this.shouldStop() && (fetchNotifs = this.fetchNotifs()) != null) {
                TargetedNotification[] targetedNotifications = fetchNotifs.getTargetedNotifications();
                int length = targetedNotifications.length;
                long lng = 0L;
                Object o;
                Integer access$1200;
                synchronized (ClientNotifForwarder.this) {
                    if (ClientNotifForwarder.this.clientSequenceNumber >= 0L) {
                        lng = fetchNotifs.getEarliestSequenceNumber() - ClientNotifForwarder.this.clientSequenceNumber;
                    }
                    ClientNotifForwarder.this.clientSequenceNumber = fetchNotifs.getNextSequenceNumber();
                    o = new HashMap();
                    for (TargetedNotification targetedNotification : targetedNotifications) {
                        Integer listenerID = targetedNotification.getListenerID();
                        if (!listenerID.equals(ClientNotifForwarder.this.mbeanRemovedNotifID)) {
                            ClientListenerInfo clientListenerInfo = (ClientListenerInfo)ClientNotifForwarder.this.infoList.get(listenerID);
                            if (clientListenerInfo != null) {
                                o.put(listenerID, clientListenerInfo);
                            }
                        }
                        else {
                            Notification notification = targetedNotification.getNotification();
                            if (notification instanceof MBeanServerNotification && notification.getType().equals("JMX.mbean.unregistered")) {
                                ClientNotifForwarder.this.removeNotificationListener(((MBeanServerNotification)notification).getMBeanName());
                            }
                        }
                    }
                    access$1200 = ClientNotifForwarder.this.mbeanRemovedNotifID;
                }
                if (lng > 0L) {
                    String string = "May have lost up to " + lng + " notification" + ((lng == 1L) ? "" : "s");
                    ClientNotifForwarder.this.lostNotifs(string, lng);
                    ClientNotifForwarder.logger.trace("NotifFetcher.run", string);
                }
                for (int j = 0; j < length; ++j) {
                    this.dispatchNotification(targetedNotifications[j], access$1200, o);
                }
            }
            synchronized (ClientNotifForwarder.this) {
                ClientNotifForwarder.this.currentFetchThread = null;
            }
            if (fetchNotifs == null && ClientNotifForwarder.logger.traceOn()) {
                ClientNotifForwarder.logger.trace("NotifFetcher-run", "Recieved null object as notifs, stops fetching because the notification server is terminated.");
            }
            if (fetchNotifs == null || this.shouldStop()) {
                ClientNotifForwarder.access$1000(ClientNotifForwarder.this, 3);
                try {
                    ClientNotifForwarder.this.removeListenerForMBeanRemovedNotif(ClientNotifForwarder.this.mbeanRemovedNotifID);
                }
                catch (Exception t) {
                    if (ClientNotifForwarder.logger.traceOn()) {
                        ClientNotifForwarder.logger.trace("NotifFetcher-run", "removeListenerForMBeanRemovedNotif", t);
                    }
                }
            }
            else {
                ClientNotifForwarder.this.executor.execute(this);
            }
        }
        
        void dispatchNotification(TargetedNotification targetedNotification, Integer obj, Map<Integer, ClientListenerInfo> map) {
            Notification notification = targetedNotification.getNotification();
            Integer listenerID = targetedNotification.getListenerID();
            if (listenerID.equals(obj)) {
                return;
            }
            ClientListenerInfo clientListenerInfo = (ClientListenerInfo)map.get(listenerID);
            if (clientListenerInfo == null) {
                ClientNotifForwarder.logger.trace("NotifFetcher.dispatch", "Listener ID not in map");
                return;
            }
            NotificationListener listener = clientListenerInfo.getListener();
            Object handback = clientListenerInfo.getHandback();
            try {
                listener.handleNotification(notification, handback);
            }
            catch (RuntimeException t) {
                ClientNotifForwarder.logger.trace("NotifFetcher-run", "Failed to forward a notification to a listener", t);
            }
            return;
        }
        
        private NotificationResult fetchNotifs() {
            try {
                NotificationResult fetchNotifs = ClientNotifForwarder.this.fetchNotifs(ClientNotifForwarder.this.clientSequenceNumber, ClientNotifForwarder.this.maxNotifications, ClientNotifForwarder.this.timeout);
                if (ClientNotifForwarder.logger.traceOn()) {
                    ClientNotifForwarder.logger.trace("NotifFetcher-run", "Got notifications from the server: " + fetchNotifs);
                    return fetchNotifs;
                }
                return fetchNotifs;
            }
            catch (ClassNotFoundException | NotSerializableException | UnmarshalException t) {
                ClientNotifForwarder.logger.trace("NotifFetcher.fetchNotifs", t);
                return this.fetchOneNotif();
            }
            catch (IOException t2) {
                if (!this.shouldStop()) {
                    ClientNotifForwarder.logger.error("NotifFetcher-run", "Failed to fetch notification, stopping thread. Error is: " + t2, t2);
                    ClientNotifForwarder.logger.debug("NotifFetcher-run", t2);
                }
                return null;
            }
        }
        
        private NotificationResult fetchOneNotif() {
            ClientNotifForwarder this$0 = ClientNotifForwarder.this;
            long n = ClientNotifForwarder.this.clientSequenceNumber;
            int i = 0;
            Object fetchNotifs = null;
            long earliestSequenceNumber = -1L;
            while (fetchNotifs == null && !this.shouldStop()) {
                NotificationResult fetchNotifs2;
                try {
                    fetchNotifs2 = this$0.fetchNotifs(n, 0, 0L);
                }
                catch (ClassNotFoundException ex) {
                    ClientNotifForwarder.logger.warning("NotifFetcher.fetchOneNotif", "Impossible exception: " + ex);
                    ClientNotifForwarder.logger.debug("NotifFetcher.fetchOneNotif", ex);
                    return null;
                }
                catch (IOException t) {
                    if (!this.shouldStop()) {
                        ClientNotifForwarder.logger.trace("NotifFetcher.fetchOneNotif", t);
                    }
                    return null;
                }
                if (this.shouldStop() || fetchNotifs2 == null) {
                    return null;
                }
                n = fetchNotifs2.getNextSequenceNumber();
                if (earliestSequenceNumber < 0L) {
                    earliestSequenceNumber = fetchNotifs2.getEarliestSequenceNumber();
                }
                try {
                    fetchNotifs = this$0.fetchNotifs(n, 1, 0L);
                }
                catch (ClassNotFoundException | NotSerializableException | UnmarshalException t2) {
                    ClientNotifForwarder.logger.warning("NotifFetcher.fetchOneNotif", "Failed to deserialize a notification: " + t2.toString());
                    if (ClientNotifForwarder.logger.traceOn()) {
                        ClientNotifForwarder.logger.trace("NotifFetcher.fetchOneNotif", "Failed to deserialize a notification.", t2);
                    }
                    ++i;
                    ++n;
                }
                catch (Exception t3) {
                    if (!this.shouldStop()) {
                        ClientNotifForwarder.logger.trace("NotifFetcher.fetchOneNotif", t3);
                    }
                    return null;
                }
            }
            if (i > 0) {
                ClientNotifForwarder.this.lostNotifs("Dropped " + i + " notification" + ((i == 1) ? "" : "s") + " because classes were missing locally or incompatible", (long)i);
                if (fetchNotifs != null) {
                    fetchNotifs = new NotificationResult(earliestSequenceNumber, fetchNotifs.getNextSequenceNumber(), fetchNotifs.getTargetedNotifications());
                }
            }
            return fetchNotifs;
        }
        
        private boolean shouldStop() {
            synchronized (ClientNotifForwarder.this) {
                if (ClientNotifForwarder.this.state != 1) {
                    return true;
                }
                if (ClientNotifForwarder.this.infoList.size() == 0) {
                    ClientNotifForwarder.access$1000(ClientNotifForwarder.this, 2);
                    return true;
                }
                return false;
            }
        }
        
        static void access$600(ClientNotifForwarder.NotifFetcher notifFetcher, String s, SecurityException ex) {
            notifFetcher.logOnce(s, ex);
        }
    }
}
